from typing import Dict, Union

from pydantic import BaseModel, Field, validator

from ..accessors import GenericImageDataAccessor
from .router import router
from .segment_zproj import segment_zproj_pipeline
from .shared import call_pipeline
from ..roiset import get_label_ids, RoiSet, RoiSetMetaParams, RoiSetExportParams
from ..session import session

from ..pipelines.shared import PipelineTrace, PipelineParams, PipelineRecord

from ..models import Model, InstanceSegmentationModel


class RoiSetObjectMapParams(PipelineParams):
    class _SegmentationParams(BaseModel):
        channel: int = Field(
            None,
            description='Channel of input image to use for solving segmentation; use all channels if empty'
        )
        zi: Union[int, None] = Field(
            None,
            description='z coordinate to use on input image when solving segmentation; apply MIP if empty',
        )

    accessor_id: str = Field(
        description='ID(s) of previously loaded accessor(s) to use as pipeline input'
    )
    pixel_classifier_segmentation_model_id: str = Field(
        description='Pixel classifier applied to segmentation_channel(s) to segment objects'
    )
    object_classifier_model_id: Union[str, None] = Field(
        None,
        description='Object classifier used to classify segmented objectss'
    )
    pixel_classifier_derived_model_id: Union[str, None] = Field(
        None,
        description='Pixel classifier used to derive channel(s) as additional inputs to object classification'
    )
    patches_channel: int = Field(
        description='Channel of input image used in patches sent to object classifier'
    )
    segmentation: _SegmentationParams = Field(
        _SegmentationParams(),
        description='Parameters used to solve segmentation'
    )
    roi_params: RoiSetMetaParams = RoiSetMetaParams(**{
        'mask_type': 'boxes',
        'filters': {
            'area': {'min': 1e3, 'max': 1e8}
        },
        'expand_box_by': [128, 2],
        'deproject_channel': None,
    })
    export_params: RoiSetExportParams = RoiSetExportParams()
    derived_channels_input_channel: Union[int, None] = Field(
        None,
        description='Channel of input image from which to compute derived channels; use all if empty'
    )
    derived_channels_output_channels: Union[int, list] = Field(
        None,
        description='Derived channels to send to object classifier; use all if empty'
    )
    export_label_interm: bool = False


class RoiSetToObjectMapRecord(PipelineRecord):
    roiset_table: dict

@router.put('/roiset_to_obmap/infer')
def roiset_object_map(p: RoiSetObjectMapParams) -> RoiSetToObjectMapRecord:
    """
    Compute a RoiSet from 2d segmentation, apply to z-stack, and optionally apply object classification.
    """
    record, rois = call_pipeline(roiset_object_map_pipeline, p)

    table = rois.get_serializable_dataframe()

    session.write_to_table('RoiSet', {'input_filename': p.accessor_id}, table)
    ret = RoiSetToObjectMapRecord(
        roiset_table=table.to_dict(),
        **record.dict()
    )
    return ret


def roiset_object_map_pipeline(
        accessors: Dict[str, GenericImageDataAccessor],
        models: Dict[str, Model],
        **k
) -> (PipelineTrace, RoiSet):
    d = PipelineTrace(accessors['accessor'])

    d['mask'] = segment_zproj_pipeline(
        accessors,
        {'model': models['pixel_classifier_segmentation_model']},
        **k['segmentation'],
    ).last

    d['labeled'] = get_label_ids(d.last)
    rois = RoiSet.from_object_ids(d['input'], d['labeled'], RoiSetMetaParams(**k['roi_params']))

    # optionally append RoiSet products
    for ki, vi in rois.get_export_product_accessors(k['patches_channel'], RoiSetExportParams(**k['export_params'])).items():
        d[ki] = vi

    # optionally run an object classifier if specified
    if obmod := models.get('object_classifier_model'):
        obmod_name = k['object_classifier_model_id']
        assert isinstance(obmod, InstanceSegmentationModel)
        rois.classify_by(
            obmod_name,
            [k['patches_channel']],
            obmod,
        )
        d[obmod_name] = rois.get_object_class_map(obmod_name)
    else:
        d['objects_unclassified'] = d.last.apply(lambda x: ((x > 0) * 1).astype('uint16'))
    return d, rois
